from __future__ import unicode_literals
import pytest
import sys
import time
from .terminal import Terminal


def pytest_addoption(parser):
    parser.addoption(
        "--coverage", action="store_true", default=False, help="generate coverage report"
    )


@pytest.fixture(scope='session')
def radian_command(pytestconfig):
    if pytestconfig.getoption("coverage"):
        radian_command = [sys.executable, "-m", "radian", "--coverage"]
    else:
        radian_command = [sys.executable, "-m", "radian"]

    return radian_command


@pytest.fixture(scope='function')
def terminal(radian_command):
    with Terminal.open(radian_command) as t:
        yield t
        t.sendintr()
        t.write("q()\n")
        start_time = time.time()
        while t.isalive():
            if time.time() - start_time > 5:
                raise Exception("radian didn't quit cleanly")
            time.sleep(0.01)



from __future__ import unicode_literals
import os


def test_shell(terminal):
    terminal.current_line().assert_startswith("r$>")
    terminal.write(";")
    terminal.current_line().assert_startswith("#!>")
    terminal.write("python --version\n")
    terminal.previous_line(2).assert_startswith("Python ")
    terminal.current_line().assert_startswith("#!>")
    terminal.write("\b")
    terminal.current_line().assert_startswith("r$>")


def test_cd(terminal):
    terminal.current_line().assert_startswith("r$>")
    terminal.write(";")
    d = os.path.realpath(os.path.join(os.path.dirname(__file__), "..", "radi"))
    terminal.write("cd {}".format(d))
    terminal.current_line().strip().assert_endswith(os.sep + "radi")
    try:
        terminal.write("\t")
        terminal.current_line().strip().assert_endswith(os.sep + "radian")
    except Exception:
        terminal.write("\t")
        terminal.current_line().strip().assert_endswith(os.sep + "radian")
    terminal.write("\n")
    terminal.previous_line(2).strip().assert_endswith(os.sep + "radian")
    terminal.write("cd -\n")
    terminal.previous_line(2).strip().assert_equal(os.getcwd())


def test_cd2(terminal):
    terminal.current_line().assert_startswith("r$>")
    terminal.write(";")
    d = os.path.realpath(os.path.join(os.path.dirname(__file__), "..", "radi"))
    terminal.write("cd \"{}".format(d))
    terminal.current_line().strip().assert_endswith(os.sep + "radi")
    try:
        terminal.write("\t")
        terminal.current_line().strip().assert_endswith(os.sep + "radian")
    except Exception:
        terminal.write("\t")
        terminal.current_line().strip().assert_endswith(os.sep + "radian")
    terminal.write("\"\n")
    terminal.previous_line(2).strip().assert_endswith(os.sep + "radian")
    terminal.write("cd -\n")
    terminal.previous_line(2).strip().assert_equal(os.getcwd())



from __future__ import unicode_literals
import sys

from radian.settings import radian_settings as settings


TERMINAL_CURSOR_AT_BEGINNING = [True]


def create_read_console(session):
    interrupted = [False]

    def read_console(message, add_history=1):
        session.prompt_text = message

        activated = False
        for name in reversed(session.modes):
            mode = session.modes[name]
            if mode.activator and mode.activator(session):
                session.activate_mode(name)
                activated = True
                break
        if not activated:
            session.activate_mode("unknown")

        current_mode = session.current_mode

        if interrupted[0]:
            interrupted[0] = False
        elif not TERMINAL_CURSOR_AT_BEGINNING[0] or \
                (settings.insert_new_line and current_mode.insert_new_line):
            session.app.output.write("\n")

        text = None

        while text is None:
            try:
                text = session.prompt(add_history=add_history)

            except KeyboardInterrupt:
                interrupted[0] = True
                raise

            except Exception as e:
                if isinstance(e, EOFError):
                    # todo: confirmation in "r" mode
                    return None
                else:
                    print("unexpected error was caught.")
                    print("please report to https://github.com/randy3k/radian for such error.")
                    print(e)
                    import traceback
                    traceback.print_exc()
                    sys.exit(1)

            current_mode = session.current_mode

            if not current_mode.native:
                result = current_mode.on_done(session)
                if result is not None:
                    return result
                if settings.insert_new_line and current_mode.insert_new_line:
                    session.app.output.write("\n")
                text = None

        return text

    return read_console


def create_write_console_ex(session, stderr_format):
    from prompt_toolkit.formatted_text import ANSI
    from prompt_toolkit.shortcuts import print_formatted_text

    # color_depth = session.color_depth

    def write_console_ex(buf, otype):
        if otype == 0:
            if sys.stdout:
                sys.stdout.write(buf)
                sys.stdout.flush()
                TERMINAL_CURSOR_AT_BEGINNING[0] = buf.replace("\r", "\n").endswith("\n")
        else:
            if sys.stderr:
                print_formatted_text(ANSI(stderr_format.format(buf)), end="", file=sys.stderr)
                sys.stderr.flush()
                TERMINAL_CURSOR_AT_BEGINNING[0] = buf.replace("\r", "\n").endswith("\n")

    return write_console_ex



from __future__ import unicode_literals
import os
import sys
from rchitect import rcopy, rcall, reval
from rchitect._cffi import ffi, lib
from rchitect.interface import roption, protected, rstring_p
from .key_bindings import map_key


def prase_text_complete(text):
    status = ffi.new("ParseStatus[1]")
    s = rstring_p(text)
    orig_stderr = sys.stderr
    sys.stderr = None
    with protected(s):
        lib.R_ParseVector(s, -1, status, lib.R_NilValue)
        sys.stderr = orig_stderr
    return status[0] != 2


def package_is_loaded(pkg):
    return pkg in rcopy(rcall(("base", "loadedNamespaces")))


def package_is_installed(pkg):
    return pkg in rcopy(reval("rownames(installed.packages())"))


def source_file(path):
    rcall(("base", "source"), path, rcall(("base", "new.env")))


def user_path(*args):
    return os.path.join(rcopy(rcall(("base", "path.expand"), "~")), *args)


def source_radian_profile(path):
    if not path:
        if os.path.exists(".radian_profile"):
            path = ".radian_profile"
        else:
            path = user_path(".radian_profile")
    path = os.path.expanduser(path)
    if os.path.exists(path):
        source_file(path)


def load_custom_key_bindings(*args):
    esc_keymap = roption("radian.escape_key_map", [])
    for m in esc_keymap:
        map_key(("escape", m["key"]), m["value"], mode=m["mode"] if "mode" in m else "r")


def register_cleanup(cleanup):
    rcall(("base", "reg.finalizer"),
          rcall(("base", "getOption"), "rchitect.py_tools"),
          cleanup,
          onexit=True)



from __future__ import unicode_literals


def test_readline(terminal):
    # issue #106
    terminal.current_line().assert_startswith("r$>")
    terminal.write("cat('hello'); readline('> ')\n")
    terminal.previous_line(1).assert_startswith("hello")
    terminal.current_line().assert_startswith("> ")



from __future__ import unicode_literals
from prompt_toolkit.application.current import get_app
from prompt_toolkit.utils import is_windows
import errno

if not is_windows():
    from prompt_toolkit.input.vt100 import Vt100Input
    from prompt_toolkit.output.vt100 import Vt100_Output

    class CustomVt100Input(Vt100Input):
        @property
        def responds_to_cpr(self):
            return False

    class CustomVt100Output(Vt100_Output):

        def flush(self):
            # it is needed when the stdout was redirected
            # see https://github.com/Non-Contradiction/JuliaCall/issues/39
            try:
                if self._buffer:
                    data = ''.join(self._buffer)
                    if self.write_binary:
                        if hasattr(self.stdout, 'buffer'):
                            out = self.stdout.buffer  # Py3.
                        else:
                            out = self.stdout
                        out.write(data.encode(self.stdout.encoding or 'utf-8', 'replace'))
                    else:
                        self.stdout.write(data)
                    self._buffer = []
            except IOError as e:
                if e.args and e.args[0] == errno.EINTR:
                    pass
                elif e.args and e.args[0] == 0:
                    pass
                else:
                    raise
            try:
                self.stdout.flush()
            except IOError as e:
                if e.args and e.args[0] == errno.EAGAIN:
                    app = get_app()
                    app.renderer.render(app, app.layout)
                    pass
                else:
                    raise
else:

    class CustomVt100Input(object):
        pass

    class CustomVt100Output(object):
        pass



from __future__ import unicode_literals
import sys
import pyte
import operator
import threading
from contextlib import contextmanager
import six
import time

if sys.platform.startswith("win"):
    import winpty
else:
    import ptyprocess


__all__ = ["PtyProcess", "Screen", "ByteStream", "Terminal", "LazyTerminal"]


if sys.platform.startswith("win"):
    ParentPtyProcess = winpty.PtyProcess
else:
    ParentPtyProcess = ptyprocess.PtyProcess


class PtyProcess(ParentPtyProcess):

    def read(self, nbytes):
        if sys.platform.startswith("win"):
            return super(PtyProcess, self).read(nbytes).encode("utf-8")
        else:
            return super(PtyProcess, self).read(nbytes)

    def write(self, data):
        if sys.platform.startswith("win"):
            super(PtyProcess, self).write(data.decode("utf-8"))
        else:
            super(PtyProcess, self).write(data)


class Screen(pyte.Screen):

    def __init__(self, process, *args, **kwargs):
        self._process = process
        super(Screen, self).__init__(*args, **kwargs)

    def write_process_input(self, data):
        self._process.write(data.encode("utf-8"))


class ByteStream(pyte.ByteStream):

    def start_feeding(self):
        screen = self.listener
        process = screen._process

        def reader():
            while True:
                try:
                    data = process.read(1024)
                except EOFError:
                    break
                if data:
                    self.feed(data)
        t = threading.Thread(target=reader)
        t.start()


class Var(object):
    def __init__(self, getter):
        self.getter = getter

    def __getattr__(self, name):
        # fallback methods
        def _(*args, **kwargs):
            return Var(lambda: getattr(self.getter(), name)(*args, **kwargs))
        return _

    def observe(self, *args, **kwargs):
        return self.getter(*args, **kwargs)

    def _assert(self, op, operand, timeout=5):
        t = time.time()
        while time.time() - t < timeout:
            value = self.getter()
            if op(value, operand):
                break
            time.sleep(0.05)
        else:
            raise Exception("value is {}".format(value))

    def assert_startswith(self, operand, timeout=5):
        self._assert(six.text_type.startswith, operand, timeout)

    def assert_endswith(self, operand, timeout=5):
        self._assert(six.text_type.endswith, operand, timeout)

    def assert_equal(self, operand, timeout=5):
        self._assert(operator.eq, operand, timeout)

    def assert_contains(self, operand, timeout=5):
        self._assert(operator.contains, operand, timeout)


class Terminal(object):

    def __init__(self, process, screen, stream):
        self.process = process
        self.screen = screen
        self.stream = stream

    @classmethod
    @contextmanager
    def open(cls, cmd):
        process = PtyProcess.spawn(cmd)
        screen = Screen(process, 80, 24)
        stream = ByteStream(screen)
        stream.start_feeding()
        try:
            yield cls(process, screen, stream)
        finally:
            process.terminate(force=True)

    def sendintr(self):
        self.process.sendintr()

    def isalive(self):
        return self.process.isalive()

    def write(self, x):
        self.process.write(x.encode('utf-8'))

    def _line(self, num=0):
        # parent's `line` method
        return self.screen.display[num]

    def line(self, num=0):
        return Var(lambda: self._line(num))

    def cursor(self):
        return Var(lambda: (self.screen.cursor.x, self.screen.cursor.y))

    def current_line(self):
        return Var(lambda: self._line(self.screen.cursor.y))

    def previous_line(self, num=1):
        return Var(lambda: self._line(self.screen.cursor.y - num))



from __future__ import unicode_literals
from prompt_toolkit.completion import Completer, Completion
import os
import sys
import shlex
import re

from rchitect.interface import rcall, reval, rcopy, rstring, rint

from six import text_type


LIBRARY_PATTERN = re.compile(r"(?:library|require)\([\"']?(.*)$")

CompleteCode = """
local(suppressWarnings({{tryCatch(
    {{
        if ({settimelimit}) base::setTimeLimit({timeout})
        utils:::.completeToken()
        if ({settimelimit}) base::setTimeLimit()
    }},
    error = function(e) {{
        if ({settimelimit}) base::setTimeLimit()
        assign("comps", NULL, env = utils:::.CompletionEnv)
    }}
)}}))
"""


class RCompleter(Completer):
    initialized = False

    def __init__(self, timeout=0.02):
        self.timeout = timeout
        super(RCompleter, self).__init__()

    def get_completions(self, document, complete_event):
        token = ""
        text = document.current_line_before_cursor
        text_after = document.text_after_cursor

        completions = []
        rcall(reval("utils:::.assignLinebuffer"), rstring(text))
        rcall(reval("utils:::.assignEnd"), rint(len(text)))
        token = rcopy(text_type, rcall(reval("utils:::.guessTokenFromLine")))
        completion_requested = complete_event.completion_requested
        completions = []

        if (len(token) >= 3 and text[-1].isalnum()) or completion_requested:
            orig_stderr = sys.stderr
            sys.stderr = None
            try:
                reval(CompleteCode.format(
                    settimelimit="TRUE" if not completion_requested and self.timeout > 0 else "FALSE",
                    timeout=str(self.timeout)))
            except Exception:
                return
            finally:
                sys.stderr = orig_stderr

            completions = rcopy(list, rcall(reval("utils:::.retrieveCompletions")))
            if not completions:
                completions = []

        for c in completions:
            if c.startswith(token):
                yield Completion(c, -len(token))

        library_prefix = LIBRARY_PATTERN.match(text)
        if token and not library_prefix:
            if (len(token) >= 3 and text[-1].isalnum()) or completion_requested:
                instring = text_after.startswith("'") or text_after.startswith('"')
                packages = rcopy(list, reval("""
                    tryCatch(
                        base::rownames(utils::installed.packages()),
                        error = function(e) character(0)
                    )
                    """))
                for p in packages:
                    if p.startswith(token):
                        comp = p if instring else p + "::"
                        if comp not in completions:
                            yield Completion(comp, -len(token))


class SmartPathCompleter(Completer):
    def get_completions(self, document, complete_event):
        text = document.text_before_cursor
        if len(text) == 0:
            return

        # do not auto complete when typing
        if not complete_event.completion_requested:
            return

        if sys.platform.startswith('win'):
            text = text.replace("\\", "/")

        directories_only = False
        quoted = False

        if text.lstrip().startswith("cd "):
            directories_only = True
            text = text.lstrip()[3:]

        try:
            path = ""
            while not path and text:
                quoted = False
                try:
                    if text.startswith('"'):
                        path = shlex.split(text + "\"")[-1]
                        quoted = True
                    elif text.startswith("'"):
                        path = shlex.split(text + "'")[-1]
                        quoted = True
                    else:
                        path = shlex.split(text)[-1]
                except RuntimeError:
                    pass
                finally:
                    if not path:
                        text = text[1:]

            path = os.path.expanduser(path)
            path = os.path.expandvars(path)
            if not os.path.isabs(path):
                path = os.path.join(os.getcwd(), path)
            basename = os.path.basename(path)
            dirname = os.path.dirname(path)

            for c in os.listdir(dirname):
                if directories_only and not os.path.isdir(os.path.join(dirname, c)):
                    continue
                if c.lower().startswith(basename.lower()):
                    if sys.platform.startswith('win') or quoted:
                        yield Completion(text_type(c), -len(basename))
                    else:
                        yield Completion(text_type(c.replace(" ", "\\ ")), -len(basename))

        except Exception:
            pass



from __future__ import unicode_literals


def test_startup(terminal):
    terminal.line(0).assert_startswith("R ")
    terminal.cursor().assert_equal((4, 3))
    terminal.current_line().assert_startswith("r$>")
    terminal.write("\n")
    terminal.current_line().assert_startswith("r$>")
    terminal.cursor().assert_equal((4, 5))
    terminal.sendintr()
    terminal.current_line().assert_startswith("r$>")
    terminal.cursor().assert_equal((4, 7))
